import yaml
import numpy as np
from pathlib import Path
from typing import Optional
import warnings

import hostphot

hostphot_path = Path(hostphot.__path__[0])
filters_file = hostphot_path.joinpath("filters", "filters.yml")

# surveys that need background subtraction
bkg_surveys = [
    "2MASS",
    "WISE",
    "VISTA",
    "SkyMapper",
    "UKIDSS",
]
# surveys which are flipped respect to most others
flipped_surveys = [
    "DES",
    "VISTA",
    "UKIDSS",
]
# add warnings for the users
survey_warnings = {
    "SkyMapper": "WARNING: The photometric calibration of SkyMapper is not trustworthy at the moment (DR4)!",
    "VISTA": "WARNING: The photometric calibration of VISTA is currently not correct (might be a HostPhot problem)!",
    "UKIDSS": "WARNING: The photometric calibration of UKIDSS is currently under revision",
}


def load_yml(file: str) -> dict:
    """Simply loads a YAML file."""
    with open(file) as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def check_survey_validity(survey: str) -> None:
    """Check whether the given survey is whithin the valid
    options.

    Parameters
    ----------
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.
    """
    filters_config = load_yml(filters_file)
    surveys = list(filters_config.keys())
    assert survey in surveys, f"Survey '{survey}' not in {surveys}"
    if survey in survey_warnings.keys():
        warnings.warn(survey_warnings[survey])


def get_survey_filters(survey: str) -> str | list:
    """Gets all the valid filters for the given survey.

    Parameters
    ----------
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.

    Returns
    -------
    filters: Filters for the given survey.
    """
    check_survey_validity(survey)
    if survey in ["HST", "JWST"]:
        # For HST, the filter needs to be specified
        return None
    filters_config = load_yml(filters_file)
    filters = list(filters_config[survey].keys())
    return filters


def survey_zp(survey: str, filt: str) -> str | dict:
    """Returns the zero-point for a given survey.

    Parameters
    ----------
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.

    Returns
    -------
    zp_dict: Zero-points for all the filters in the given survey. If the survey zero-point
             is different for each image, the string ``header`` is returned.
    """
    check_survey_validity(survey)
    check_filters_validity([filt], survey)
    filters_config = load_yml(filters_file)
    zp = filters_config[survey][filt]["zeropoint"]
    return zp


def survey_pixel_scale(survey: str, filt: Optional[str] = None) -> float:
    """Returns the pixel scale for a given survey.

    Parameters
    ----------
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.
    filt: Filter to use by surveys that have different pixel scales
          for different filters (e.g. Spitzer's IRAC and MIPS instruments).

    Returns
    -------
    pixel_scale: Pixel scale in units of arcsec/pixel.
    """
    check_survey_validity(survey)
    filters_config = load_yml(filters_file)
    pixel_scale = float(filters_config[survey][filt]["pixel_scale"])
    return pixel_scale


def survey_pixel_units(survey: str, filt: str) -> str:
    """Returns the pixel units for a given survey.

    Parameters
    ----------
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.

    Returns
    -------
    pixel_units: Pixel units (e.g. ``counts`` or ``counts/second``).
    """
    check_survey_validity(survey)
    check_filters_validity([filt], survey)
    filters_config = load_yml(filters_file)
    pixel_units = filters_config[survey][filt]["pixel_units"]
    return pixel_units


def check_filters_validity(filters: str | list, survey: str) -> None:
    """Check whether the given filters are whithin the valid
    options for the given survey.

    Parameters
    ----------
    filters: Filters to use, e,g, ``griz``.
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.
    """
    if survey == "HST":
        if isinstance(filters, list):
            for filt in filters:
                check_HST_filters(filt)
        else:
            check_HST_filters(filters)
    elif survey == "JWST":
        if isinstance(filters, list):
            for filt in filters:
                check_JWST_filters(filt)
        else:
            check_JWST_filters(filters)
    else:
        valid_filters = get_survey_filters(survey)
        for filt in filters:
            message = (
                f"filter '{filt}' is not a valid option for "
                f"'{survey}' survey ({valid_filters})"
            )
            assert filt in valid_filters, message


def check_HST_filters(filt: str) -> None:
    """Check whether the given filter is whithin the valid
    options for HST.

    Parameters
    ----------
    filt: Filter to use, e,g, ``WFC3_UVIS_F225W``.
    """
    if filt is None:
        raise ValueError(f"'{filt}' is not a valid HST filter.")
    # For UVIS, only the filters of UVIS1 are used as the
    # detector 2 is scaled to match detector 1
    detector = filt.split("_")[1]
    if detector == "UVIS":
        filt_ = filt.replace("UVIS", "UVIS1")
    else:
        filt_ = filt
    filters_dir = hostphot_path / "filters/HST"
    hst_filters = [file.name.split(".")[0] for file in filters_dir.rglob("*.dat")]
    assert filt_ in hst_filters, f"Not a valid HST filter ({filt_}): {hst_filters}"


def check_JWST_filters(filt: str) -> None:
    """Check whether the given filter is within the valid
    options for JWST.

    Parameters
    ----------
    filt: str
        Filter to use, e,g, ``NIRCam_F150W``.
    """
    if filt is None:
        raise ValueError(f"'{filt}' is not a valid JWST filter.")
    filters_dir = hostphot_path / "filters/JWST"
    jwst_filters = [file.name.split(".")[0] for file in filters_dir.rglob("*.dat")]
    assert filt in jwst_filters, f"Not a valid JWST filter ({filt}): {jwst_filters}"


def extract_filter(
    filt: str, survey: str, version: Optional[str] = None
) -> tuple[np.array, np.array]:
    """Extracts the transmission function for the filter.

    Parameters
    ----------
    filt: Filter to extract.
    survey: Survey of the filters.
    version: Version of the filters to use. E.g. for the
             Legacy Survey as it uses DECam for the south
             and BASS+MzLS for the north.

    Returns
    -------
    wave: Wavelength range of the filter.
    transmission: Transmission function.
    """
    check_survey_validity(survey)
    if survey in ["HST", "JWST"]:
        check_filters_validity(filt, survey)
    else:
        check_filters_validity([filt], survey)
    if "WISE" in survey:
        survey = "WISE"  # for unWISE to use the same filters as WISE
    filters_path = Path(hostphot_path, "filters", survey)
    # Assume DECaLS filters below 32 degrees and BASS+MzLS above
    # https://www.legacysurvey.org/status/
    if survey == "LegacySurvey":
        if version == "BASS+MzLS":
            if filt == "z":
                filt_file = filters_path / "MzLS_z.dat"
            else:
                filt_file = filters_path / f"BASS_{filt}.dat"
        elif version == "DECam":
            filt_file = filters_path / f"DECAM_{filt}.dat"
    elif survey == "HST":
        if "UVIS" in filt:
            # Usually UVIS2 is used, but there is no large difference
            filt_ = filt.replace("UVIS", "UVIS2")
        else:
            filt_ = filt
        filt_file = [
            file for file in Path(filters_path, "HST").rglob("*.dat") if filt_ in file
        ][0]
    elif survey == "JWST":
        detector = filt.split("_")[0]
        filt_file = filters_path / f"{detector}/{filt}.dat"
    else:
        filt_file = filters_path / f"{survey}_{filt}.dat"
    wave, transmission = np.loadtxt(filt_file).T
    return wave, transmission



# -*- coding: utf-8 -*-

import os
from dotenv import load_dotenv
from matplotlib.font_manager import findSystemFonts

load_dotenv()
workdir = os.getenv("workdir", "images")

# for plots
font_family = 'serif'
font_families = findSystemFonts(fontpaths=None, fontext='ttf')
for family in font_families:
    if 'P052' in family:
        font_family = "P052"
        break



import tarfile
import requests
import pandas as pd
from pathlib import Path
from typing import Optional

import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astroquery.skyview import SkyView
from astropy.coordinates import SkyCoord

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning


def get_WISE_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None) -> list[fits.ImageHDU]:
    """Downloads a set of WISE fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses all WISE filters.

    Return
    ------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    survey = "WISE"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    pixel_scale = survey_pixel_scale(survey, "W1")  # same pixel scale for all filters
    image_size = size_arcsec / pixel_scale
    coords = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")

    base_url = "https://irsa.ipac.caltech.edu/SIA?COLLECTION=wise_allwise"
    params_url = f"&POS=circle+{ra}+{dec}+0.002777&RESPONSEFORMAT=CSV&FORMAT=image/fits"
    response = requests.get(base_url + params_url)
    # transform the response into a dataframe
    split_text = response.text.split("\n")
    url_dict = {key:[] for key in split_text[0].split(",")}
    for row in split_text[1:]:
        if row == "":
            continue
        values = row.split(",")
        for key, value in zip(url_dict.keys(), values):
            url_dict[key].append(value)
    url_df = pd.DataFrame(url_dict)

    hdu_list = []
    for filt in filters:
        filt_df = url_df[url_df.energy_bandpassname==filt]
        if len(filt_df) == 0:
            hdu_list.append(None)
        else:
            img_url = filt_df.access_url.values[0]
            t_expt = filt_df.t_exptime.values[0]
            hdu = fits.open(img_url)
            hdu[0].header["EXPTIME"] = t_expt
            # create cutout
            wcs = WCS(hdu[0].header)
            cutout = Cutout2D(hdu[0].data, coords, image_size, wcs=wcs)
            hdu[0].data = cutout.data
            hdu[0].header.update(cutout.wcs.to_header())
            hdu_list.append(hdu)
    return hdu_list

def get_unWISE_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None, version: str = "allwise") -> list[fits.ImageHDU]:
    """Downloads a set of unWISE fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses all WISE filters.
    version:  Version of the unWISE images. Either ``allwise`` or ``neo{i}`` for {i} = 1 to 7.

    Return
    ------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    survey = "unWISE"
    if version is None:
        version = "allwise"
    else:
        # check validity of the version used
        neo_versions = [f"neo{i}" for i in range(1, 8)]
        all_versions = ["allwise"] + neo_versions
        assert (
            version in all_versions
        ), f"Not a valid version ({version}): {all_versions}"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if "neo" in version:
        # these only have W1 and W2 data
        if "W3" in filters:
            filters.remove("W3")
        if "W4" in filters:
            filters.remove("W4")

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    pixel_scale = survey_pixel_scale(survey, "W1")  # same pixel scale for all filters
    size_pixels = int(size_arcsec.value / pixel_scale)
    assert size_pixels <= 1024, "Maximum cutout size for unWISE is 1024 pixels"

    bands = "".join(filt[-1] for filt in filters)  # e.g. 1234
    # for more info: http://unwise.me/imgsearch/
    base_url = "http://unwise.me/cutout_fits?"
    params_url = (
        f"version={version}&ra={ra}&dec={dec}&size={size_pixels}&bands={bands}"
    )
    master_url = base_url + params_url

    response = requests.get(master_url, stream=True)
    target_file = Path(f"unWISE_images_{ra}_{dec}.tar.gz")  # current directory
    if response.status_code == 200:
        with open(target_file, "wb") as f:
            f.write(response.raw.read())

    hdu_list = []
    with tarfile.open(target_file) as tar_file:
        files_list = tar_file.getnames()
        for fits_file in files_list:
            for filt in filters:
                if f"{filt.lower()}-img-m.fits" in fits_file:
                    tar_file.extraction_filter = (lambda member, path: member)
                    tar_file.extract(fits_file, ".")
                    hdu = fits.open(fits_file)
                    hdu_list.append(hdu)
                    Path(fits_file).unlink()  # remove file
    # remove the tarfile
    if target_file.is_file() is True:
        target_file.unlink()
    return hdu_list

def get_used_image(header: fits.header.Header) -> str:
    """Obtains the name of the image downloaded by SkyView.

    Parameters
    ----------
    header: fits header
        Header of an image.

    Returns
    -------
    used_image: str
        Name of the image.
    """
    image_line = header["HISTORY"][-3]
    used_image = image_line.split("/")[-1].split("-")[0]

    return used_image

def _get_WISE_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None) -> list[fits.ImageHDU]:
    """Downloads a set of WISE fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses all WISE filters.

    Return
    ------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    survey = "WISE"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        # just to get the original image used
        skyview_fits = SkyView.get_images(
            position=coords,
            coordinates="icrs",
            pixels="100",
            survey="WISE 3.4",
        )
        header = skyview_fits[0][0].header
        coadd_id = get_used_image(header)
        coadd_id1 = coadd_id[:4]
        coadd_id2 = coadd_id1[:2]

        # for more info: https://irsa.ipac.caltech.edu/ibe/docs/wise/allwise/p3am_cdd/#sample_code
        base_url = (
            "http://irsa.ipac.caltech.edu/ibe/data/wise/allwise/p3am_cdd/"
        )
        coadd_url = Path(coadd_id2, coadd_id1, coadd_id)
        params_url = f"center={ra},{dec}&size={size_arcsec.value}arcsec&gzip=0"  # center and size of the image
        hdu_list = []
        for filt in filters:
            i = filt[-1]
            band_url = f"{coadd_id}-w{i}-int-3.fits"
            url = Path(
                base_url, coadd_url, band_url + "?" + params_url
            )
            print(url)
            hdu = fits.open(url)
            hdu_list.append(hdu)
    return hdu_list


import numpy as np
from astropy import wcs
from astropy.io import fits
from astropy.coordinates import SkyCoord
from astropy.nddata.utils import Cutout2D

import warnings
from astropy.utils.exceptions import AstropyWarning


def trim_images(
    fits_files: list[fits.ImageHDU], pos: SkyCoord, size: float
) -> list[fits.ImageHDU]:
    """Trims the size of the given fits images.

    Parameters
    ----------
    fits_files: List of fits images.
    pos: The position of the cutout array's center.
    size: Image size in pixels.

    Returns
    -------
    trimmed_fits_files: List of the trimmed fits images.
    """
    trimmed_fits_files = []
    for fits_file in fits_files:
        data = fits_file[0].data.copy()
        header = fits_file[0].header
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = wcs.WCS(header, naxis=2)

        # trim data and update the header with the WCS
        trimmed_data = Cutout2D(data, pos, size, img_wcs, copy=True)
        fits_file[0].data = trimmed_data.data
        header.update(trimmed_data.wcs.to_header())
        header["COMMENT"] = "= Trimmed fits file (hostphot)"
        fits_file[0].header = header
        trimmed_fits_files.append(fits_file)

    return trimmed_fits_files


def remove_nan(hdu: list[fits.ImageHDU]) -> list[fits.ImageHDU]:
    """Remove columns and/or rows which have all NaN values.
    The WCS is updated accordingly.

    Parameters
    ----------
    image: Header Data Unit
        Fits image with header and data.

    Returns
    -------
    trimmed_hdu: Trimmed image.
    """
    header = hdu[0].header
    data = hdu[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    mask = ~np.all(np.isnan(data), axis=0)
    x_size = np.sum(np.ones_like(data[0])[mask])
    x_pos = np.median(np.arange(data.shape[1])[mask])

    mask = ~np.all(np.isnan(data), axis=1)
    y_size = np.sum(np.ones_like(data[:, 0])[mask])
    y_pos = np.median(np.arange(data.shape[0])[mask])

    # trim data and update the header with the WCS
    trimmed_data = Cutout2D(
        data,
        (x_pos, y_pos),
        (y_size, x_size),  # has to be (ny, nx)
        wcs=img_wcs,
        copy=True,
    )
    header.update(trimmed_data.wcs.to_header())
    header["COMMENT"] = "= Trimmed fits file (hostphot)"

    trimmed_hdu = hdu.copy()
    trimmed_hdu[0].data = trimmed_data.data
    trimmed_hdu[0].header = header

    return trimmed_hdu



import numpy as np
import pandas as pd
from typing import Optional

from pyvo.dal import sia
from astropy.io import fits
from astropy import units as u

from hostphot.surveys_utils import get_survey_filters, check_filters_validity


def get_DES_urls(ra: float, dec: float, fov: float, filters: str="grizY") -> tuple[list[str], list[str]]:
    """Obtains the URLs of the DES images+weights with the
    largest exposure times in the given filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    fov: Field of view in degrees.
    filters: Filters to uses.

    Returns
    -------
    url_list:  List of URLs with DES images.
    url_w_list: List of URLs with DES images weights.
    """
    # check filters
    if filters is None:
        filters = get_survey_filters("DES")
    check_filters_validity(filters, "DES")
    # get field-of-view table
    des_access_url = "https://datalab.noirlab.edu/sia/des_dr1"
    svc = sia.SIAService(des_access_url)
    imgs_table = svc.search(
        (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
    )
    if len(imgs_table) == 0:
        print(("Warning: empty table returned for " f"ra={ra}, dec={dec}"))
        return None, None
    # conditions to select images of interest
    imgs_df = pd.DataFrame(imgs_table)
    cond_stack = imgs_df["proctype"] == "Stack"
    cond_image = imgs_df["prodtype"] == "image"
    cond_weight = imgs_df["prodtype"] == "weight"

    url_list = []
    url_w_list = []
    for filt in filters:
        # gather images of interes
        cond_filt = imgs_df["obs_bandpass"] == filt
        selected_images = imgs_df[cond_filt & (cond_stack & cond_image)]
        selected_weights = imgs_df[cond_filt & (cond_stack & cond_weight)]
        if len(selected_images) > 0:
            # pick image with the longest exposure time
            max_exptime = np.argmax(selected_images["exptime"])
            row = selected_images.iloc[max_exptime]
            url = row["access_url"]  # get the download URL
            if len(selected_weights) == 0:
                url_w = None
            else:
                max_exptime = np.argmax(selected_weights["exptime"])
                row_w = selected_weights.iloc[max_exptime]
                url_w = row_w["access_url"]
        else:
            url = url_w = None
        url_list.append(url)
        url_w_list.append(url_w)
    return url_list, url_w_list

def get_DES_images(ra: float, dec: float, size: float | u.Quantity = 3, filters: Optional[str] = None) -> list[fits.HDUList] | None:
    """Gets DES fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``grizY``.

    Returns
    -------
    hdu_list: List of fits images. ``None`` if the images are not found.
    """
    # check filters
    survey = "DES"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    # get image size in degrees
    if isinstance(size, (float, int)):
        fov = (size * u.arcmin).to(u.degree).value
    else:
        fov = size.to(u.degree).value
    # get URLs
    url_list, url_w_list = get_DES_urls(ra, dec, fov, filters)
    if url_list is None:
        return None
    # download images
    hdu_list = []
    for url, url_w in zip(url_list, url_w_list):
        # combine image+weights on a single fits file
        image_hdu = fits.open(url)
        hdu = fits.PrimaryHDU(image_hdu[0].data, header=image_hdu[0].header)
        if url_w is None:
            hdu_sublist = fits.HDUList([hdu])
        else:
            weight_hdu = fits.open(url_w)
            hdu_err = fits.ImageHDU(
                weight_hdu[0].data, header=weight_hdu[0].header
            )
            hdu_sublist = fits.HDUList([hdu, hdu_err])
            weight_hdu.close()
        hdu_list.append(hdu_sublist)
        image_hdu.close()
    return hdu_list


import numpy as np
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt

from astropy import wcs
from astropy.io import fits
from photutils.aperture import EllipticalAperture

import hostphot
from hostphot.surveys_utils import check_survey_validity


hostphot_path = Path(hostphot.__path__[0])
config_file = hostphot_path.joinpath("filters", "config.txt")
config_df = pd.read_csv(config_file, sep="\\s+")
plt.rcParams["mathtext.fontset"] = "cm"


def pixel2pixel(
    x1: float, y1: float, img_wcs1: wcs.WCS, img_wcs2: wcs.WCS
) -> tuple[float, float]:
    """Convert the pixel coordinates from one image to another."""
    coord1 = img_wcs1.pixel_to_world(x1, y1)
    x2, y2 = img_wcs2.world_to_pixel(coord1)

    return x2, y2


def get_image_gain(header: fits.Header, survey: str) -> float:
    """Returns the gain from an image's header.

    **Note:** for ``SDSS`` this is assumed to be one
    as it should already be included.

    Parameters
    ----------
    header: Header of an image.
    survey: Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    gain: Gain value.
    """
    check_survey_validity(survey)
    if survey == "PS1":
        gain = header["HIERARCH CELL.GAIN"]
    elif survey == "DES":
        gain = header["GAIN"]
    elif survey == "SDSS":
        gain = 1.0
    elif survey == "2MASS":
        # the value comes from https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        # Note that it is different from the value of
        # https://iopscience.iop.org/article/10.1086/498708/pdf (8 e ADU^-1)
        gain = 10.0
    elif survey == "LegacySurvey":
        gain = 1  # not used
    elif survey == "Spitzer":
        # also in the "EXPGAIN" keyword
        if header["INSTRUME"] == "IRAC":
            # Table 2.4 of IRAC Instrument Handbook
            gain = 3.8  # all filters have similar gain
            gain *= header["EFCONV"]  # convert DN/s to MJy/sr
        elif header["INSTRUME"] == "MIPS":
            # Table 2.4 of MIPS Instrument Handbook
            gain = 5.0
    elif survey == "VISTA":
        # use median from http://casu.ast.cam.ac.uk/surveys-projects/vista/technical/vista-gain
        gain = 4.19
    elif survey == "HST":
        gain = header["CCDGAIN"]
    elif survey == "SkyMapper":
        gain = header["GAIN"]
    elif survey == "SPLUS":
        gain = header["GAIN"]
    elif survey == "UKIDSS":
        gain = header["GAIN"]
    else:
        gain = 1.0

    return gain


def get_image_readnoise(header: fits.Header, survey: str) -> float:
    """Returns the read noise from an image's header.
    All values are per-pixel values.

    **Note:** for ``SDSS`` this is assumed to be zero
    as it should already be included.

    Parameters
    ----------
    header: Header of an image.
    survey: Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    readnoise: Read noise value.
    """
    check_survey_validity(survey)
    if survey == "PanSTARRS":
        readnoise = header["HIERARCH CELL.READNOISE"]
    elif survey == "DES":
        # see https://arxiv.org/pdf/0810.3600.pdf
        readnoise = 7.0  # electrons per pixel
    elif survey == "SDSS":
        readnoise = 0.0
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        # 6 combined images
        readnoise = 4.5 * np.sqrt(6)  # not used
    elif survey == "LegacySurvey":
        readnoise = 1.0  # not used
    elif survey == "Spitzer":
        if header["INSTRUME"] == "IRAC":
            # Table 2.3 of IRAC Instrument Handbook
            # very rough average
            readnoise_dict = {1: 16.0, 2: 12.0, 3: 10.0, 4: 8.0}
            channel = header["CHNLNUM"]
            readnoise = readnoise_dict[channel]
        elif header["INSTRUME"] == "MIPS":
            # Table 2.4 of MIPS Instrument Handbook
            readnoise = 40.0
    elif survey == "VISTA":
        # very rough average for all filters in
        # http://casu.ast.cam.ac.uk/surveys-projects/vista/technical/vista-gain
        readnoise = 24.0
    elif survey == "HST":
        # tipically 0.0
        readnoise = header["PCTERNOI"]
    elif survey == "SkyMapper":
        # https://rsaa.anu.edu.au/observatories/instruments/skymapper-instrument
        readnoise = 5  # electrons
    elif survey == "SPLUS":
        readnoise = header["HIERARCH OAJ QC NCNOISE"]
    elif survey == "UKIDSS":
        readnoise = header["READNOIS"]
    else:
        readnoise = 0.0

    return readnoise


def get_image_exptime(header: fits.Header, survey: str) -> float:
    """Returns the exposure time from an image's header.

    Parameters
    ----------
    header: Header of an image.
    survey: Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    exptime: Exposure time in seconds.
    """
    check_survey_validity(survey)
    if survey in ["PanSTARRS", "DES", "SDSS", "GALEX", "VISTA", "Spitzer", "HST", "SkyMapper"]:
        exptime = float(header["EXPTIME"])
    elif survey == "WISE":
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/
        # and https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec1_1.html
        if header["BAND"] in [1, 2]:
            exptime = 7.7
        elif header["BAND"] in [3, 4]:
            exptime = 8.8
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        exptime = 7.8
    elif survey == "LegacySurvey":
        exptime = 1.0  # ???
    elif survey == "SPLUS":
        exptime = header["TEXPOSED"]
    elif survey == "UKIDSS":
        exptime = header["EXP_TIME"] * header["NEXP"]
    elif survey == "JWST":
        exptime = header["XPOSURE"]
    else:
        exptime = 1.0

    return exptime


def correct_HST_aperture(filt: str, ap_area: float, header: fits.Header) -> float:
    """Get the aperture correction for the given configuration.

    see: https://hst-docs.stsci.edu/wfc3dhb/chapter-9-wfc3-data-analysis/9-1-photometry
         https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints
         https://www.stsci.edu/hst/instrumentation/wfc3/data-analysis/photometric-calibration/uvis-photometric-calibration
    
    Parameters
    ----------
    filt: HST filter, e.g. ``WFC3_UVIS_F225W``.
    ap_area: Aperture area.
    header: Header of an image.

    Returns
    -------
    correction: Aperture correction (encircled energy fraction).
    """
    # split instrument and filter
    filt_split = filt.split("_")
    filt = filt_split[-1]
    instrument = filt_split[-2]

    if instrument == "UVIS":
        # either UVIS1 or UVIS2
        instrument = header["APERTURE"]
        # some images have a different APERTURE value
        # see: https://hst-docs.stsci.edu/wfc3ihb/chapter-6-uvis-imaging-with-wfc3/6-4-uvis-field-geometry
        # not sure if this is the correct solution
        if instrument == "UVIS-CENTER":
            instrument = "UVIS2"
        if instrument == "UVIS":
            instrument = "UVIS1"

    # assuming circular aperture
    # for an ellipse, this would take the average of the axes
    ap_radius = np.sqrt(ap_area / np.pi)

    # get correction curve
    ac_file = [
        file
        for file in Path(hostphot_path, "filters", "HST").glob(
            f"*{instrument.lower()}_aper*"
        )
    ][0]
    ac_df = pd.read_csv(ac_file)

    # linear interpolation of the aperture correction
    apertures = np.array(
        [
            float(col.replace("APER#", ""))
            for col in ac_df.columns
            if col.startswith("AP")
        ]
    )
    ap_corr = ac_df[ac_df.FILTER == filt].values[0][2:].astype(float)

    cont_apertures = np.arange(0, 9, 0.01)
    cont_ap_corr = np.interp(cont_apertures, apertures, ap_corr)

    # get the closest value
    ind = np.argmin(np.abs(cont_apertures - ap_radius))
    correction = cont_ap_corr[ind]

    return correction


def adapt_aperture(
    objects: np.ndarray, img_wcs: wcs.WCS, img_wcs2: wcs.WCS, flip: bool = False
) -> tuple[np.ndarray, float]:
    """Changes the aperture parameters to consider differences
    in WCS between surveys.

    The values of ``center``, ``a`` and ``b`` should in pixel
    units. DES images are flipped, so these need to be corrected
    with ``theta -> -theta``, i.e. using ``flip=True``.

    Parameters
    ----------
    objects: Objects with apertures.
    img_wcs: WCS of the image from where the objects were extracted.
    img_wcs2: WCS used to adapt the apertures.
    flip: Whether to flip the orientation of the aperture. Only
        used for DES images.

    Returns
    -------
    objects_: Objects with adapted apertures.
    conv_factor: Convertion factor between the resolutions of the images.
    """
    objects_ = objects.copy()  # avoid modifying the intial objects
    for obj in objects_:
        center = (obj["x"], obj["y"])
        apertures = EllipticalAperture(center, obj["a"], obj["b"], obj["theta"])
        sky_apertures = apertures.to_sky(img_wcs)

        new_apertures = sky_apertures.to_pixel(img_wcs2)
        obj["x"], obj["y"] = new_apertures.positions
        obj["a"] = new_apertures.a
        obj["b"] = new_apertures.b
        obj["theta"] = new_apertures.theta

        if flip is True:
            # flip aperture orientation
            obj["theta"] *= -1

    # ratio between ellipse axis of two images
    conv_factor = np.mean(np.copy(objects["a"] / objects_["a"]))

    return objects_, conv_factor



import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional
from copy import deepcopy
import matplotlib.pyplot as plt
import aplpy

import sep
from astropy.io import fits
from astropy import wcs
from astropy.coordinates import concatenate
from astropy.convolution import (
    Gaussian2DKernel,
    convolve_fft,
    interpolate_replace_nans,
)

from hostphot._constants import workdir, font_family
from .cleaning import remove_nan
from .objects_detection import (
    extract_objects,
    find_gaia_objects,
    find_catalog_objects,
    cross_match,
)
from hostphot.utils import suppress_stdout, store_input
from hostphot.photometry.image_utils import adapt_aperture
from hostphot.surveys_utils import (
    check_survey_validity,
    bkg_surveys,
    flipped_surveys,
)

import warnings
from astropy.utils.exceptions import AstropyWarning

plt.rcParams["mathtext.fontset"] = "cm"


def mask_image(
    data: np.ndarray,
    objects: np.ndarray,
    r: float | np.ndarray = 6,
    sigma: float | np.ndarray = 8,
) -> np.ndarray:
    """Masks objects in an image (2D array) by convolving it with
    a 2D Gaussian kernel.

    Parameters
    ----------
    data: Image data.
    objects: Objects extracted with :func:`sep.extract()`.
    r: Scale of the semi-mayor and semi-minor axes
        of the ellipse of the `objects`.
    sigma: Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.

    Returns
    -------
    masked_data: Masked image data.
    """
    mask = np.zeros(data.shape, dtype=bool)
    sep.mask_ellipse(
        mask,
        objects["x"],
        objects["y"],
        objects["a"],
        objects["b"],
        objects["theta"],
        r=r,
    )

    masked_data = data.copy()
    masked_data[mask] = np.nan
    # mask data by convolving it with a 2D Gaussian kernel
    # with the same sigma in x and y
    kernel = Gaussian2DKernel(sigma)
    masked_data = interpolate_replace_nans(masked_data, kernel, convolve_fft)

    return masked_data


def create_mask(
    name: str,
    host_ra: float,
    host_dec: float,
    filt: str,
    survey: str,
    ra: Optional[float] = None,
    dec: Optional[float] = None,
    bkg_sub: Optional[bool] = None,
    threshold: float = 15,
    sigma: float | np.ndarray = 8,
    r: float | np.ndarray = 6,
    crossmatch: bool = False,
    gal_dist_thresh: float = -1,
    deblend_cont: float = 0.005,
    ref_filt: Optional[str] = None,
    ref_survey: Optional[str] = None,
    save_plots: bool = False,
    save_mask_params: bool = False,
    save_input: bool = True,
) -> None:
    """Calculates the aperture parameters to mask detected sources.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    host_ra: Host-galaxy right ascension in degrees.
    host_dec: Host-galaxy declination in degrees.
    filt: Filter to use to load the fits file. List is commonly used for coadds.
    survey: Survey to use for the zero-points and correct filter path.
    ra: Right ascension of an object, in degrees. Used for plotting.
    dec: Declination of an object, in degrees. Used for plotting.
    bkg_sub: If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: Threshold used by :func:`sep.extract()` to extract objects.
    sigma: Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.
    r: Scale of the aperture size for the sources to be masked.
    crossmatch: If ``True``, the detected objects are cross-matched with a
        Gaia catalog.
    crossmatch: If ``True``, the detected objects are cross-matched with a
        Gaia catalog.
    gal_dist_thresh: Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    deblend_cont : Minimum contrast ratio used for object deblending. Default is 0.005.
        To entirely disable deblending, set to 1.0.
    ref_filt: Reference filter (or coadd filters) from which to use the mask parameters. Note that the parameters
        need to have been previously saved with ``save_mask_params=True``.
    ref_survey: Reference survey from which to use the mask parameters. Note that the parameters
        need to have been previously saved with ``save_mask_params=True``.
    save_plots: If ``True``, the mask and galaxy aperture figures are saved.
    save_mask_params: If `True`, the extracted mask parameters are saved for later use.
    save_input: Whether to save the input parameters.
    """
    input_params = locals()  # dictionary
    check_survey_validity(survey)
    if isinstance(filt, list):
        filt = "".join(f for f in filt)

    obj_dir = Path(workdir, name)
    fits_file = obj_dir / survey / f"{survey}_{filt}.fits"
    hdu = fits.open(fits_file)
    hdu = remove_nan(hdu)

    header = hdu[0].header
    data = hdu[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    # save input parameters
    if save_input is True:
        inputs_file = obj_dir / survey / "input_masking_parameters.csv"
        store_input(input_params, inputs_file)

    if (ref_filt is None) & (ref_survey is None):
        # extract objects
        gal_obj, nongal_objs = extract_objects(
            data_sub,
            bkg_rms,
            host_ra,
            host_dec,
            threshold,
            img_wcs,
            gal_dist_thresh,
            deblend_cont,
        )
        # preprocessing: cross-match extracted objects with a catalog
        # using two Gaia catalogs as they do not always include the
        # same objects.
        if crossmatch:
            cat_coord1 = find_gaia_objects(host_ra, host_dec)
            cat_coord2 = find_catalog_objects(host_ra, host_dec)
            cat_coord = concatenate([cat_coord1, cat_coord2])
            nongal_objs = cross_match(nongal_objs, img_wcs, cat_coord)
    else:
        # use objects previously extracted
        # the aperture/ellipse parameters are updated accordingly
        gal_obj, nongal_objs, master_img_wcs, sigma, r, flip2 = load_mask_params(
            name, ref_filt, ref_survey
        )

        if survey != ref_survey:
            # cross-survey mask: need to adapt some values
            # flipping images
            if survey in flipped_surveys:
                flip = True
            else:
                flip = False
            if flip == flip2:
                flip_ = False
            else:
                flip_ = True
            # adapt sizes
            gal_obj, _ = adapt_aperture(gal_obj, master_img_wcs, img_wcs, flip_)
            nongal_objs, conv_factor = adapt_aperture(
                nongal_objs, master_img_wcs, img_wcs, flip_
            )
            sigma /= conv_factor

    #masked_data = mask_image(data_sub, nongal_objs, r=r, sigma=sigma)
    if len(nongal_objs) > 0:
        masked_data = mask_image(data_sub, nongal_objs, r=r, sigma=sigma)
    else:
        masked_data = data_sub.copy()
    masked_hdu = deepcopy(hdu)
    masked_hdu[0].data = masked_data
    outfile = obj_dir / survey / f"{survey}_{filt}_masked.fits"
    masked_hdu.writeto(outfile, overwrite=True)

    if survey in flipped_surveys:
        flip = True
    else:
        flip = False

    if save_mask_params is True:
        # save detected objects and masking parameters
        objects_df = pd.concat([pd.DataFrame(gal_obj), pd.DataFrame(nongal_objs)])
        objects_df["sigma"] = sigma
        objects_df["r"] = r
        objects_df["flip"] = flip
        objects_df["filt"] = filt
        objects_df["survey"] = survey
        outfile = obj_dir / survey / f"mask_parameters_{filt}.csv"
        objects_df.to_csv(outfile, index=False)

    if save_plots is True:
        outfile = obj_dir / survey / f"{survey}_{filt}_masked.jpg"
        title = fr"{name}: {survey}-${filt}$"
        plot_masked_image(
            hdu,
            masked_hdu,
            gal_obj,
            nongal_objs,
            r,
            host_ra,
            host_dec,
            ra,
            dec,
            title,
            outfile,
        )
    hdu.close()


def load_mask_params(
    name: str, filt: str | list, survey: str
) -> tuple[np.ndarray, np.ndarray, wcs.WCS, np.ndarray, float, bool]:
    """Loads previously saved mask parameters.

    Parameters
    ----------
    name: Name of the object to find the path of the mask-parameters file.
    filt: Name of the filter used to create the mask parameters. Coadds are
        also valid.
    survey: Survey name to be used.

    Returns
    -------
    gal_obj, nongal_objs, img_wcs, sigma, r, flip: Mask parameters.
    """
    if isinstance(filt, list):
        filt = ''.join(f for f in filt)
    obj_dir = Path(workdir, name)
    mask_params_file = obj_dir / survey / f"mask_parameters_{filt}.csv"
    objects_df = pd.read_csv(mask_params_file)
    
    # split parameters
    sigma = objects_df.pop("sigma").values[0]
    r = objects_df.pop("r").values[1:]  # remove host-galaxy row
    flip = objects_df.pop("flip").values[0]
    # remove unused columns
    _ = objects_df.pop("filt")
    _ = objects_df.pop("survey")
    # DataFrame to structured/record array
    objects = objects_df.to_records()
    gal_obj = objects[:1]
    nongal_objs = objects[1:]
    
    # load image WCS
    fits_file = obj_dir / survey / f"{survey}_{filt}.fits"
    hdu = fits.open(fits_file)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(hdu[0].header, naxis=2)
        
    return gal_obj, nongal_objs, img_wcs, sigma, r, flip


def plot_masked_image(
    hdu: list[fits.ImageHDU],
    masked_hdu: list[fits.ImageHDU],
    gal_obj: np.ndarray,
    objects: np.ndarray,
    r: float | np.ndarray = 6,
    host_ra: Optional[float] = None,
    host_dec: Optional[float] = None,
    ra: Optional[float] = None,
    dec: Optional[float] = None,
    title: Optional[str] = None,
    outfile: Optional[float] = None,
) -> None:
    """Plots the masked image together with the original image and
    the detected objects.

    Parameters
    ----------
    data: Image FITS file.
    masked_data: Masked image FITS file.
    objects: Objects extracted with :func:`sep.extract()`.
    img_wcs: Image's WCS.
    gal_obj: Galaxy object.
    r: Scale of the aperture size for the masked sources.
    host_ra: Right ascension of the galaxy, in degrees. Used for plotting the position of the galaxy.
    host_dec: Declination of the galaxy, in degrees. Used for plotting the position of the galaxy.
    ra: Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: Declination of an object, in degrees. Used for plotting the position of the object.
    title: Title of the image.
    outfile: If given, path where to save the output figure.
    """
    figure = plt.figure(figsize=(30, 12))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig1 = aplpy.FITSFigure(hdu, figure=figure, subplot=(1, 3, 1))
        fig2 = aplpy.FITSFigure(hdu, figure=figure, subplot=(1, 3, 2))
        fig3 = aplpy.FITSFigure(masked_hdu, figure=figure, subplot=(1, 3, 3))
    fig2.tick_labels.hide_y()
    fig2.axis_labels.hide_y()
    fig3.tick_labels.set_yposition("right")
    fig3.axis_labels.set_yposition("right")

    figures = [fig1, fig2, fig3]
    titles = ["Initial Image", "Detected Sources", "Masked Image"]
    for i, fig in enumerate(figures):
        fig.set_theme("publication")
        fig.set_title(titles[i], **{"family": font_family, "size": 24})
        with suppress_stdout():
            fig.show_grayscale(stretch="arcsinh")
        # ticks
        fig.tick_labels.set_font(**{"family": font_family, "size": 18})
        fig.tick_labels.set_xformat("dd.dd")
        fig.tick_labels.set_yformat("dd.dd")
        fig.ticks.set_length(6)
        # ToDo: solve this deprecation warning (Aplpy should do it?)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            fig.axis_labels.set_font(**{"family": font_family, "size": 18})  # this line is giving a warning!
    # galaxy markers
    fig2.show_markers(
        host_ra,
        host_dec,
        edgecolor="k",
        facecolor="m",
        alpha=0.7,
        marker="P",
        s=250,
        label="Given galaxy",
    )
    fig2.show_markers(
        gal_obj["x"][0],
        gal_obj["y"][0],
        edgecolor="k",
        facecolor="r",
        alpha=0.7,
        marker="X",
        s=250,
        label="Identified galaxy",
        coords_frame="pixel",
    )
    # galaxy pseudo-aperture
    """
    fig2.show_ellipses(
        gal_obj["x"][0],
        gal_obj["y"][0],
        2 * r * gal_obj["a"][0],
        2 * r * gal_obj["b"][0],
        gal_obj["theta"][0] * 180.0 / np.pi,
        coords_frame="pixel",
        linewidth=3,
        edgecolor="r",
        linestyle="dotted",
    )
    """
    # other sources apertures
    fig2.show_ellipses(
        objects["x"],
        objects["y"],
        2 * r * objects["a"],
        2 * r * objects["b"],
        objects["theta"] * 180.0 / np.pi,
        coords_frame="pixel",
        linewidth=2,
        edgecolor="orangered",
        linestyle="dotted",
    )
    # SN marker
    if (ra is not None) and (dec is not None):
        for fig in figures[1:]:
            fig.show_markers(
                ra,
                dec,
                edgecolor="k",
                facecolor="aqua",
                marker="*",
                s=250,
                label="SN",
            )

    fig2.ax.legend(
        fancybox=True, framealpha=1, prop={"size": 20, "family": font_family}
    )
    # show sources indeces
    for i, (x, y) in enumerate(zip(objects["x"], objects["y"])):
        # the host galaxy has index one
        fig2.ax.text(x, y, i+2, fontsize=14, color="orangered")
    # title
    if title is not None:
        length = len(title) - 2
        text = fig1.ax.text(
            0.022 * length,
            0.06,
            title,
            fontsize=28,
            horizontalalignment="center",
            verticalalignment="center",
            transform=fig1.ax.transAxes,
            font=font_family,
        )
        text.set_bbox(
            dict(facecolor="white", edgecolor="white", alpha=0.9, boxstyle="round")
        )
    if outfile:
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(figure)
    else:
        plt.show()
